#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define NO_FASTLOAD

#ifdef __riscos__
  #include "S2T_RO.h"
#else
/* Include your equivalent to stat() here, and #define if required. */
  #include <sys/stat.h>
  #include <unistd.h>
#endif

#ifndef HAVE_STRICMP
#include <ctype.h>
int stricmp (const char *s1, const char *s2)
{
  char c1, c2;
  int d;
  do {
    c1 = toupper (*s1++);
    c2 = toupper (*s2++);
    d = c2 - c1;
    if (d < 0) return -1;
    if (d > 0) return 1;
  } while (c1 && c2);
  return 0;
}
#endif


/* Assumes ASCII character set */

#define MAXBLOCKS 12

typedef const char cchar;

int	NumBlocks,
	verbose=0,
	fast=0,
	BlockDataLen[MAXBLOCKS],
	BlockParam1 [MAXBLOCKS],
	BlockParam2 [MAXBLOCKS],
	BlockLen    [MAXBLOCKS],
	BlockType   [MAXBLOCKS];

char	*BlockStart  [MAXBLOCKS],
	BlockFlag   [MAXBLOCKS],
	BlockName   [MAXBLOCKS][11];


int fail(int err)
{
  static const char *msgs[] = {
    NULL,
    "Input file is not of type Z80 or SNA",
    "Usage: snatotap <snapshot> <TAP file> [<loader name>]",
    "Not enough memory!"
  };

#define FAIL_badInput	1
#define FAIL_Usage	2
#define FAIL_noMem	3

  fprintf(stderr, "%s\n", msgs[err]);
  return err;
}


#ifdef __riscos__

#define fatal(f,d) ro_fatal(d)

void ro_fatal(FILE *fd)
{
  char msg[256];

  strcpy(msg, _kernel_last_oserror()->errmess);
  fprintf(stderr, "%s\n", msg);
  setvar(msg);
  if (fd!=NULL) fclose(fd);
  exit(255);
}

#else

void fatal(cchar *file, FILE *fd)
{
  perror(file);
  if (fd!=NULL) fclose(fd);
  exit(255);
}

#endif

void SaveAsTape(cchar* file)
{
  FILE *fh1;
  int n;
  char buff[2];

  if (verbose) puts("Writing TAP file...");
  fh1=fopen(file,"wb");
  if (fh1==NULL) fatal(file, NULL);
  for (n=0; n<NumBlocks; n++) {
    buff[0]=(char) (BlockLen[n]&0xff);
    buff[1]=(char) (BlockLen[n]>>8);
    if (fwrite(buff, 2, 1, fh1)!=1) fatal(file, fh1);
    if (fwrite(BlockStart[n], BlockLen[n], 1, fh1)!=1) fatal(file, fh1);
  }
  if (fclose(fh1)) fatal(file, NULL);
#ifdef __riscos__
  {
    _kernel_oserror *p;
    if (verbose) puts("Setting filetype...");
    p=_setType(file, 0x11D);
    if (p) { memcpy(_kernel_last_oserror(), p, 256); fatal(file, NULL); }
  }
#endif
}


char SlowLoad[21+153]={
 19,  0,  0,  0, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32,149,  0,  2,  0,149,
  0, 88,151,  0,255,  0,  1,127,  0,234,  1,254, 26, 33,  0, 64, 62,  0,119,
 35, 11,184, 32,250,185, 32,247, 54, 13, 35, 54, 13, 33,127,254, 34,254, 81,
 33, 72,  2, 34,254, 82, 33,  8,248, 34,254, 83, 33,  8,128, 34,254, 84, 33,
  8,130, 34,254, 85, 33, 62,254, 34,254, 86, 17, 76,255, 33, 37, 93,  1, 41,
  0,237,176, 49,  0,  0,221, 33,  0, 64, 17, 76,191, 62,153, 55,195, 76,255,
205, 86,  5, 49,224, 87,193,209,225,241,221,225,253,225,217,  8,241,237, 71,
241,237, 79,241,211,254, 49,226,  0, 17, 76,255, 33,224, 80,  6,  0, 62,  6,
195,224, 86,  0,  2, 14,  0,249,192, 50, 51, 55, 54, 48, 14,  0,  0,208, 92,
  0, 13, 66 };

/* This is the code (first block) for the 128k loader */

char L128[634]={
 19,  0,  0,  0, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 97,  2,  2,  0, 97,
  2, 51, 99,  2,255,  0,  1, 75,  2,234,243, 49,  0,192, 33,226, 92, 17,  0,
128,  1, 55,  2,237,176,195,  0,128,  1,253,127, 58, 35,130,237,121, 33,  0,
 64, 17,  1, 64,  1,255, 26, 54,  0,237,176,205,232,129,221, 42, 11,130,221,
229, 17,  0, 64, 62,153, 55, 20,  8, 21, 62, 15,211,254,205, 98,  5,221,225,
205,165,129, 33,  0,192, 17,  0, 64,  1,  0, 64,237,176,  6,  6, 33, 13,130,
 17, 28,130,197,  1,253,127, 26,237,121, 19,213, 94, 35, 86, 35,237, 83, 53,
130,221, 42, 53,130,229,221,229, 17,  0, 64, 62,153, 55, 20,  8, 21, 62, 15,
211,254,205, 98,  5,221,225,205,165,129,225,209,193, 16,206, 33,224, 90, 17,
225, 90,  1, 29,  0, 54,255,237,176,205,232,129, 33,209,128, 17,224, 80,  6,
  7,197,  1, 30,  0,237,176,  1,226,  0,213,221,225,221,  9,221,229,209,193,
 16,236, 49,254, 87,  1,253,127, 58, 34,130,237,121,221, 42, 25,130, 17,  0,
192,213,221,229, 17,  0, 64, 62,153, 55, 20,  8, 21, 62, 15,211,254, 58, 36,
130,203, 71,194,224, 80,195,243, 83,205, 98,  5,221,225,209,  6,  0,221,126,
  0,254,237, 40,  7, 18,221, 35, 19,195,246, 81,221,126,  1,254,237,195,224,
 81, 62,237,194,239, 80,213,225,221,126,  3,119,221, 78,  2, 13, 19,237,176,
 14,  4,221,  9,122,167,194,232, 80,195,224, 82, 49,224, 85, 22, 15,  1,253,
255, 21,237, 81,241,  1,253,191,237,121,122,167, 32,240, 49,224, 86,193,209,
225,195,224, 83,241,221,225,253,225,217,  8,241,237, 71,241,237, 79,241,211,
254,195,224, 84,205, 98,  5,195,224, 82,  0,  0,  0,  0,  0,  1,253,255, 62,
255,237,121,  1,253,127, 62,255,237,121,193,209,225,241, 49,  0,255,237, 70,
243,195,  0,255,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 17,  0,192,  6,  0, 58, 36,
130,203, 71, 32,  6,203, 63, 50, 36,130,201,203, 63, 50, 36,130,221,126,  0,
254,237, 40,  6, 18,221, 35, 19, 24, 26,221,126,  1,254,237, 62,237, 32,241,
213,225,221,126,  3,119,221, 78,  2, 13, 19,237,176, 14,  4,221,  9,122,167,
 32,213,201, 33,254, 90, 54, 13, 35, 54, 13,  6,  8, 33, 37,130,221, 33,254,
 80, 86,221,114,  0, 35, 86,221,114,  1, 35, 17,  0,  1,221, 25, 16,239,201,
  1,  0,  2,  0,  3,  0,  4,  0,  5,  0,  6,  0,  7,  0,  8,  0, 21, 16, 17,
 19, 20, 22, 23, 18, 16,255,  0,  0,127,254, 72,  2,  8,248,  8,128,  8,130,
 62,254,  0,  0,  0,  0, 13,  0,  2, 14,  0,249,192, 50, 51, 55, 54, 48, 14,
  0,  0,208, 92,  0, 13,226};


/*
 * FastLoad doesn't work for me ... dont know WHAT could be wrong though...
 * Speccy just stops loading after it gets the Sync bit ... [Tomaz]
 */

#ifndef NO_FASTLOAD

char FastLoad[21+214]={
 19,  0,  0,  0, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32,210,  0,  2,  0,210,
  0, 86,212,  0,255,  0,  1,188,  0,234,  1,254, 26, 33,  0, 64, 62,  0,119,
 35, 11,184, 32,250,185, 32,247, 54, 13, 35, 54, 13, 33,127,254, 34,254, 81,
 33, 72,  2, 34,254, 82, 33,  8,248, 34,254, 83, 33,  8,128, 34,254, 84, 33,
  8,130, 34,254, 85, 33, 62,254, 34,254, 86, 17, 76,255, 33, 37, 93,  1,101,
  0,237,176, 49,  0,  0,221, 33,  0, 64, 17, 76,191, 62,153, 55,195, 76,255,
205,117,255, 49,224, 87,193,209,225,241,221,225,253,225,217,  8,241,237, 71,
241,237, 79,241,211,254, 49,226,  0, 17, 76,255, 33,224, 80,  6,  0, 62,  6,
195,224, 86,243,213, 17,  0,  0,205, 98,  5,209,205,147,255,208,221,117,  0,
221, 35, 27,122,179, 32,242,205,147,
/*255,208,195,223,5,46,1,6,177,205,168,255,208,62,203,184,203,21,6,176,*/
255,208,195,223,  5, 46,  1,  6,177,205,227,  5,208, 62,203,184,203, 21,  6,
176, 48,243,125,172,103,201,205,172,255,208, 62, 22,195,233,  5, 13,  0,  2,
 14,  0,249,192, 50, 51, 55, 54, 48, 14,  0,  0,208, 92,  0, 13, 24 };

#endif

char ScreenReg[16]={
	0x0e,0x1e,0xed,0xb0,0x39,0x3d,0x20,0xf8,
	0x31,0xf2,0x57,0xc1,0xd1,0xe1,0xf1,0x31};


void DecompressZ80(char *s, char *d, int l)
{
/*
 * Decompresses a portion of .Z80 file pointed to by s to d, the destination
 * length is in l
 */
  int n=0, m=0, o;

  while (n<l) {
    if (s[m]==0xED && s[m+1]==0xED) {
      for (o=s[m+2]; o; --o)
	d[n++]=s[m+3];
      m+=4;
    } else
      d[n++]=s[m++];
  }
}

int TestDecZ80(char *s, char *d, int l, int ll)
{
/*
 * Decompresses a portion of .Z80 file pointed to by s to d, the destination
 * length is in l. Return 1 if it can stay decompressed (it never overlaps
 * with the source).
 * ll is the length of source; this is needed because the source AND
 * destination are in the same 16k block when loading to the Spectrum (at the
 * 49152 address) - though it almost never stays decompressed; only in
 * special cases
 */
  int n=0, m=0, o, over=1;

  while (n<l) {
    if (s[m]==0xED && s[m+1]==0xED) {
      for (o=s[m+2]; o; --o)
	d[n++]=s[m+3];
      m+=4;
    } else
      d[n++]=s[m++];
    if ((l-ll+m)<n) over=0;
  }
  return over;
}

int FillSNA(char *s, char *d)
{
/*
 * Copies the snapshot content from s to d and fills the appropriate
 * addresses with register values (for SNA files)
 */

  signed int n, i;
  char *p;
  unsigned short t;

  memcpy(d, s+27, 49152);
    /* Copy the snapshot contents to destination */

  p=d+6112;
  p[ 0]=s[ 5]; p[ 1]=s[ 6];	/* B'C' */
  p[ 2]=s[ 3]; p[ 3]=s[ 4];	/* D'E' */
  p[ 4]=s[ 1]; p[ 5]=s[ 2];	/* H'L' */
  p[ 6]=s[ 7]; p[ 7]=s[ 8];	/* A'F' */
  p[ 8]=s[17]; p[ 9]=s[18];	/* IX */
  p[10]=s[15]; p[11]=s[16];	/* IY */
  p[12]=0    ; p[13]=s[ 0];	/* I */
  p[14]=0    ; p[15]=s[20];	/* R */
  p[16]=0    ; p[17]=s[26]&7;	/* BORDER */
  p[18]=s[13]; p[19]=s[14];	/* BC */
  p[20]=s[11]; p[21]=s[12];	/* DE */
  p[22]=s[ 9]; p[23]=s[10];	/* HL */
  p[24]=s[21]; p[25]=s[22];	/* AF */
  p=d+5856;
  t=s[23] | s[24]<<8;		/* SP = SP+2 */
  p[16]=(char) ((t+2) & 0xff);	/* (PC is on */
  p[17]=(char) ((t+2) >> 8);	/*  stack !) */
  switch (s[25] & 3) {
    case 0: p[19]=0x46; break;	/* IM 0 */
    case 1: p[19]=0x56; break;	/* IM 1 */
    default: p[19]=0x5e;	/* IM 2 */
  }
  p[20]=(s[19]&4) ? 0xFB : 0xF3;/* EI or DI */
  p[22]=d[t-16384];		/* PC */
  p[23]=d[t-16383];		/* ( on Stack ) */
  return 0;
}


int FillZ80(char *s, char *d)
{
/*
 * Decompresses the .Z80 file from s to d and fills the appropriate addresses
 * with register values
 */
  int i, l;
  signed int n;
  char *p;

  if (!s[6] && !s[7]) {
/* 2.0+ .Z80 File */
    if (!s[34] || s[34]==1) {
      p=s+(s[30] | s[31]<<8)+32;
      for (n=3; n; --n) {
	l=p[0] | p[1]<<8;
	switch (p[2]) {
	  case 4: DecompressZ80(&p[3],&d[16384],16384); break;
	  case 5: DecompressZ80(&p[3],&d[32768],16384); break;
	  case 8: DecompressZ80(&p[3],&d[00000],16384); break;
	}
	p+=l+3;
      }
    }
    else return 1;	/* Not 48k Snapshot */
  } else {
/* 1.45 .Z80 file */
    if (s[12]&32) DecompressZ80(&s[30],d,49152);/* if compressed */
    else memcpy(d, s+30, 49152);		/* if not */
  }
/*
 * well this could be done better with a simple structure... but I didn't
 * think too much when I was working on it ;)
 */
  p=d+6112;
  p[ 0]=s[15]; p[ 1]=s[16];	/* B'C' */
  p[ 2]=s[17]; p[ 3]=s[18];	/* D'E' */
  p[ 4]=s[19]; p[ 5]=s[20];	/* H'L' */
  p[ 6]=s[22]; p[ 7]=s[21];	/* A'F' */
  p[ 8]=s[25]; p[ 9]=s[26];	/* IX */
  p[10]=s[23]; p[11]=s[24];	/* IY */
  p[12]=0    ; p[13]=s[10];	/* I */
  p[14]=0    ; p[15]=(s[11]&127 | s[12]<<7)&0xFF;/* R */
  p[16]=0    ; p[17]=(s[12]>>1)&7;/* BORDER */
  p[18]=s[ 2]; p[19]=s[ 3];	/* BC */
  p[20]=s[13]; p[21]=s[14];	/* DE */
  p[22]=s[ 4]; p[23]=s[ 5];	/* HL */
  p[24]=s[ 1]; p[25]=s[ 0];	/* AF */
  p=d+5856;
  p[16]=s[ 8]; p[17]=s[ 9];	/* SP */
  switch (s[29]&3) {
    case 0: p[19]=0x46; break;	/* IM 0 */
    case 1: p[19]=0x56; break;	/* IM 1 */
    default: p[19]=0x5e;	/* IM 2 */
  }
  p[20]=(s[27]&1) ? 0xFB : 0xF3;/* EI or DI */
  i=(s[6] | s[7]) ? 6 : 32;	/* PC (v1.54 @ offset 6, v2+ @ offset 32) */
  p[22]=s[i]; p[23]=s[i+1];	/* PC - Z80 v2.0+ */
  return 0;
}


void CalcCheckSum(int n)
{
/* Calculates Checksum for Block b */
  int l=BlockLen[n]-1;
  char c=0, *p=BlockStart[n];

  n=0;
  while (n<l) c^=p[n++];
  p[n]=c;
}


char *FillSnapData(char *Data, long Len, cchar* File, int type)
/* Data - Pointer to .SNA or .Z80 file (in memory)
 * Len  - Length of .SNA or .Z80 file
 * File - filename to use (from the Spectrum's viewpoint)
 * type - true if of type SNA
 */
{
  signed int n;
  int c, m, mp, i,l,ll, r, num, Bright;
  char *s, *d, *p, page,video, *SnaPtr;
  char BColour[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };

/*
 * NOTE:  This procedure holds the 128k CONVERTOR TOO !! I was too lazy
 *	  to make another procedure for this ... sorry
 *
 * Creates the  array ( pointed to by SnaPtr ) and then copies the Snapshot
 * pointed to by Data length Len there ( decompresses the .Z80 file and puts
 * the registers to the right place, making a .TAP block )
 *
 * Prepare Leading Program : Make Name for it and Calculate Checksum for the
 * Header, then Fill in the Data for the Data block
 * This Part is Selectable for FAST and SLOW loading !
 * FAST WILL NEVER BE IMPLEMENTED , SORRY! NO TIME ! :(
 *   I left the code in for a last chance ;))
 */

  if (verbose) puts("Building loader...");

#ifdef NO_FASTLOAD
  memcpy(SlowLoad+4, File, 10);
  memcpy(L128+4, File, 10); /* copy filename across */
#else
  s = (fast ? FastLoad : SlowLoad)+6;
  memcpy(s, File, 10);
  memcpy(L128+4, File, 10); /* copy filename across */
#endif

  NumBlocks=2;
  BlockLen[0]=19; BlockFlag[0]=0; BlockType[0]=0;
  n=0;
  BlockName[0][10]=0;
  BlockDataLen[0]=BlockParam2[0]=232; BlockParam1[0]=2; BlockFlag[1]=255;
  if (!fast) {
    BlockStart[0]=&SlowLoad[2];
    memcpy(BlockName[0], SlowLoad+6, 10);
    BlockStart[1]=&SlowLoad[23]; BlockLen[1]=151;
#ifndef NO_FASTLOAD
  } else {
    FastLoad[20]=0;
    BlockStart[0]=&FastLoad[2];
    memcpy(BlockName[0], FastLoad+6, 10);
    BlockStart[1]=&FastLoad[23]; BlockLen[1]=212;
#endif
  }
  CalcCheckSum(0);
  CalcCheckSum(1);

  if (verbose) puts("Converting snapshot...");

  SnaPtr=malloc(132000);  /* Should be enough for all the structures... */
  if (SnaPtr==NULL) return NULL;
  if (type)
    r=FillSNA(Data,&SnaPtr[3]);
  else r=FillZ80(Data,&SnaPtr[3]);
  if (verbose) printf("Snapshot is %iK\n", r ? 128 : 48);
  if (!r) {
    NumBlocks=3;
/*
 * Make entries for the TAP file and insert the actual length and type of the
 * data section
 */
    BlockStart[2]=&SnaPtr[2]; BlockLen[2]=48974; BlockFlag[2]=0x99;
    SnaPtr[0]=48974&255; SnaPtr[1]=48974>>8; SnaPtr[2]=0x99;
/* Copy mem from 65356 to 20704 ( 6 lines of 30 bytes ) */
    mp=65356+3-16384;
    i=4323;
    for (n=6; n; --n) {
      memcpy(SnaPtr+i, SnaPtr+mp, 30);
      i+=256; mp+=30;
    }
/* Copy code to fill registers after loading to video mem on 16384+5856 */
    memcpy(d=SnaPtr+3+5856, ScreenReg, 16);
    d[18]=0xed; d[21]=0xc3;
/*
 * Set the last line on screen to its most frequent background colour
 * and make it bright if it was mostly bright
 */
    if (verbose) puts("Fiddling last display line (holds loader data)");
    Bright=0;
    mp=6883;
    for (n=30; n; --n) {
      if (SnaPtr[mp] & 64) Bright++;
      BColour[(SnaPtr[mp++] & 56) >> 3]++;
    }
    for (c=n=0; n<8; ++n) if (BColour[n]>BColour[c]) c=n;
    Bright = (Bright>=15) ? 64 : 0;
    mp=6883;
    for (n=30; n; --n)
      SnaPtr[mp++]=c*9 | Bright;
/* Copy the 'TC' logo to the screen (with colours) */
    SnaPtr[mp+1]=SnaPtr[mp]=13;
    d=SnaPtr+4323+30; p=SlowLoad+52;
    d[0]=d[1]=0; d+=256; /* Erm, I seem to be thinking of ARM assembler */
    for (n=6; n; --n) {
      d[1]=p[1]; d[0]=p[0];
      d+=256; p+=6;
    }
    d[0]=d[1]=0;
    CalcCheckSum(2);
  } else {
/* 128k SNAPSHOT !!!!!!!
 * There is a BIG difference between 48k and 128k loader (apart from being
 * coded almost 5 months apart) ... the 48k has the registers and every other
 * values in the MEMORY portion of the .TAP file (3rd block) ... the loader
 * is always the same (apart from the NAME) ... the 128k file has ALL 8
 * Memory pages UNALTERED ... they are left the same as in .Z80 file
 * (only some pages must be decompressed (which must be done very rarely))
 * so the loader contains all the registers and stuff ...
 * NOTE: I left the compression as it is in .Z80 files... my decompressor
 * for the spectrum decompresses the block VERY FAST (faster than Brian's
 * loader does ;) ) ...
 */

    NumBlocks=2;
    BlockStart[0]=&L128[2]; BlockLen[0]=19; BlockFlag[0]=0; BlockType[0]=0;
    CalcCheckSum(0);
    n=0; while (n<8 && L128[6+n]!=8) { BlockName[0][n]=L128[6+n]; n++; }
    BlockName[0][n]=0;
    BlockDataLen[0]=BlockParam2[0]=232; BlockParam1[0]=2;
    BlockStart[1]=&L128[23]; BlockLen[1]=611; BlockFlag[1]=255;

    s=Data;
    NumBlocks=10;

#define OS1 26+22
#define OS2 8+22

    d=L128+OS1;
    d[547]=0;		/* Which Pages are Compressed */
    i=(s[30] | s[31]<<8)+32;
    for (n=8; n; --n) {
      static cchar pagenos[] = {8,8,8,1,2,7,3,4,0,5,6};
      l=s[i]+s[i+1]*256;
      num=pagenos[s[i+2]];
      if (TestDecZ80(&s[i+3],&SnaPtr[1+(16386*num)],16384,l)) {
        if (verbose) printf("Leaving block %i compressed\n", num);
        /* The block can stay COMPRESSED ! */
        d[547]|=(1<<num);/* Set Compressed BIT for that block */
        BlockStart[num+2]=&s[i+2];
        BlockLen[num+2]=l+2;
        ll=65536-l;
        d[522+(num<<1)]=(char)(ll & 0xff); /* Address to load to */
        d[523+(num<<1)]=(char)(ll >> 8);
        BlockFlag[num+2]=s[i+2]=0x99;
      } else {	/* The block must be loaded DECOMPRESSED ! */
        if (verbose) printf("Decompressing block %i\n", num);
        BlockStart[num+2]=&SnaPtr[16386*num];
        BlockLen[num+2]=16384+2;
        ll=65536-16384;
        d[522+(num<<1)]=(char)(ll & 0xff); /* Address to load to */
        d[523+(num<<1)]=(char)(ll >> 8);
        BlockFlag[num+2]=SnaPtr[16386*num]=0x99;
      }
      i+=l+3;
    }
    for (n=8; n; --n, CalcCheckSum(n+2));

/*
 * Set the last line on screen to its most frequent background colour
 * and make it bright if it was mostly bright
 */
    if (verbose) puts("Fiddling last display line (holds loader data)");
    Bright=0;
    mp=6881;
    for (n=30; n; --n) {
      if (SnaPtr[mp] & 64) Bright++;
      BColour[(SnaPtr[mp++] & 56) >> 3]++;
    }
    for (c=n=0; n<8; ++n) if (BColour[n]>BColour[c]) c=n;
    L128[162+15]=c*9 | ((Bright>=15) ? 64 : 0);

    d=L128+OS2;
    p=d+406;
    p[ 0]=s[15]; p[ 1]=s[16];		/* B'C' */
    p[ 2]=s[17]; p[ 3]=s[18];		/* D'E' */
    p[ 4]=s[19]; p[ 5]=s[20];		/* H'L' */
    p[ 6]=s[22]; p[ 7]=s[21];		/* A'F' */
    p[ 8]=s[25]; p[ 9]=s[26];		/* IX */
    p[10]=s[23]; p[11]=s[24];		/* IY */
    p[12]=0    ; p[13]=s[10];		/* I */
    p[14]=0    ; p[15]=(s[11]&127)|(s[12]<<7);	/* R */
    p[16]=0    ; p[17]=(s[12]>>1)&7;	/* BORDER */
    p[18]=s[ 2]; p[19]=s[ 3];		/* BC */
    p[20]=s[13]; p[21]=s[14];		/* DE */
    p[22]=s[ 4]; p[23]=s[ 5];		/* HL */
    p[24]=s[ 1]; p[25]=s[ 0];		/* AF */
    d[350]=s[38];			/* AY Reg. */
    d[357]=s[35];			/* 128 Page */
    p=d+376;
    mp=39+14;
    for (n=16; n; --n) {		/* Fill 15 AY */
      *(p++)=0; *(p++)=s[mp--];	/* Registers */
    }
    p=d+365;
    p[0]=s[8]; p[1]=s[9];		/* SP */
    switch (s[29] & 3) {
      case 0: p[3]=0x46; break;	/* IM 0 */
      case 1: p[3]=0x56; break;	/* IM 1 */
      default: p[3]=0x5e; break;	/* IM 2 */
    }
    p[4] = (s[27] & 1) ? 0xFB : 0xF3;	/* EI or DI */
    p[6]=s[32]; p[7]=s[33];		/* PC */
    CalcCheckSum(1);
  }
  return SnaPtr;
}


void showhelp()
{
  fprintf(stderr,
#ifdef __riscos__
"snatotap version 1.01, %s\n\
\n%s\
<source> must be a valid Z80 or SNA file (recognised by filetype).\n\
<dest> is the filename of the destination TAP file. Note that any existing\n\
  file of this name is overwritten.\n\
<loader name> is the name of the loader, as stored in the TAP file. If omitted,\n\
  then the leafname of the source file is used.\n\
-verbose - information about the TAP file is displayed.\n\
-fastload - a fast-load version is generated. (Not currently implemented)\n\
\n\
Freeware, Tomaz Kac and Darren Salt, 1996\n\n"
#else
"snatotap version 1.01, %s\n\
\n%s\
<source> must be a valid .Z80 or .SNA file.\n\
<dest> is the filename of the destination .TAP file. Note that any existing\n\
  file of this name is overwritten.\n\
<loader name> is the name of the loader, as stored in the TAP file. If omitted,\n\
  then the leafname (less extension) of the source file is used.\n\
-verbose - information about the TAP file is displayed.\n\
-fastload - a fast-load version is generated. (Not currently implemented)\n\
\n\
Freeware, Tomaz Kac and Darren Salt, 1996\n\n"
#endif
  ,__DATE__,
"Usage: snatotap <source> <dest> [<loader name>] [-verbose] [-fastload]\n");
  exit(0);
}


typedef union {
  struct { cchar *source, *dest, *tapname; } n;
  cchar *c[3];
} Names;


int main(int argc, cchar* argv[])
{
  char file[] = "          "; /* 10 spaces */
  int n, m, type, argp=1;
  int help, nptr=0;
  Names names;
  char *fileData, *dest;
  struct stat t;
  FILE *in;

  help=(argc==1);
  while (--argc) {
    cchar *p = argv[argp++];
    if (p[0]=='-') {
      if (!stricmp(p, "-v") || !stricmp(p, "-verbose"))
        verbose=1;
      else
      if (!stricmp(p, "-f") || !stricmp(p, "-fastload"))
        fast=1;
      else
      if (!stricmp(p, "-h") || !stricmp(p, "-help"))
        help=1;
      else return fail(FAIL_Usage);
    } else {
      if (nptr==3) return fail(FAIL_Usage);
      names.c[nptr++]=p;
    }
  }

  if (help) showhelp();

  if (nptr<2) return fail(FAIL_Usage);

  if (fast) fprintf(stderr, "Warning: -fastload specified. The fast loader option is not implemented.\n");

/* Check filetype and get file length */
#ifdef __riscos__
  n=fileType(names.n.source);
  if (n!=0xBA0 && n!=0xBA1) return fail(FAIL_badInput);
  type=(n==0xBA1);
#else
  {
    cchar *p;
    static char tmp[]="\0\0\0\0\0";
    p=strrchr(names.n.source,'.');
    if (p) {
      n=strlen(p); if (n>5) n=5;
      memcpy(tmp, p, n);
      if (stricmp(tmp, ".SNA") && stricmp(tmp, ".Z80")) p=0;
    }
    if (!p) return fail(FAIL_badInput);
    type=!stricmp(tmp, ".SNA");
  }
#endif
  if (verbose) printf("File recognised as %s\n", type ? ".SNA" : ".Z80");
  stat(names.n.source, &t);

/* Get Spectrum filename */
  if (nptr==3) {
    n=strlen(names.n.tapname); if (n>10) n=10;
    memcpy(file, names.n.tapname, n);
  } else {
#ifdef __riscos__
    cchar *p, *q;
    m=0;
    p=strrchr(names.n.source, '.');
    q=strrchr(names.n.source, ':');
    if (q>p) p=q; else if (p==NULL) p=names.n.source-1;
    while (*(++p) && m<10) file[m++]=*p;
#else
/* Modify and add to suit your system! */
    cchar *p;
    m=0;
    p=strrchr(names.n.source, '\\');
    if (p==NULL) p=names.n.source;
    while (*p && *p!='.' && m<10) { file[m++]=*p; ++p; }
#endif
    if (verbose) puts("No loader name given - using input file leafname");
  }

  if (verbose) puts("Reading snapshot...");
  fileData=malloc(t.st_size);
  if (!fileData) return fail(FAIL_noMem);

  in=fopen(names.n.source, "rb");
  if (fread(fileData, t.st_size, 1, in)!=1) fatal(argv[1], in);
  fclose(in);

  if (FillSnapData(fileData, t.st_size, file, type)==0)
    return fail(FAIL_noMem);

  SaveAsTape(names.n.dest);
#ifdef __riscos__
  setvar("");
#endif
  return 0;
}
